'''
Framework for gridding/ungridding geolocated data.

GridDef is an abstract class.  It cannot be used
as a standalone because none of the functionality
has been implemented.  It is simply a framework,
with some of the general functionality implemented.

All subclasses of GridDef provide forward and 
inverse projections, both to gridded space
and native projection space.

Subclasses MUST following the naming convention
of projectionName_GridDef for the dispatcher 
to work properly

Functions implemented for each class:
    requiredParms - the parameters required
        to be present in the dictionary argument
        passed to the constructor for the class.
        All parameters not in this list are 
        ignored.  This class MUST be made a 
        static method using the static method
        wrapper.
    indLims - return the limits of indices for
        which information should be stored.  Any
        pixels entirely outside these bounds (as 
        defined by the mapping function) will be 
        thrown out.  Note that fractional indices
        of the maxima should still be included (ie
        if maxRow is 10, data up through 10.99999
        should be be included, but not 11)
        Should return (minRow, maxRow, 
        minCol, maxCol)
    geoToProjected - from geo-coordinates (lat/lon
        in degrees N/E, respectively) to native
        projection coordinates.  Accepts numpy
        arrays.  Returns (y,x)
    geoToGridded - from geo-coordinates to 
        gridded coordinates (indices).  Indices
        must be 0-based.  Accepts numpy arrays.
        Returns (row,col) in DECIMAL INDICES.
        These must be floored before they can 
        be used as actual indices.
    projectedToGeo - from native projection 
        coordinates (y/x, usually in meters) to
        geo-coordinates. Accepts numpy arrays.
        Returns (lat,lon) in degrees.
    griddedToGeo - from decimal indices (row/col) 
        to geo-coordinates.  Accepts numpy arrays.
        Returns (lat,lon) in degrees.
'''
import sys

from pyproj import Proj

def ValidProjections():
    '''Return a list of valid projection names'''
    currentModule = sys.modules[__name__]
    names = dir(currentModule)
    return [el[:-8] for el in names if el.endswith("_GridDef")]
    
class GridDef:
    '''Abstract class to handle projection/gridding of data'''
    def __init__(self, parms):
        pass
    def requiredParms():
        raise NotImplementedError
    requiredParms = staticmethod(requiredParms)
    def indLims(self):
        raise NotImplementedError
    def geoToProjected(self, lat, lon):
        raise NotImplementedError
    def geoToGridded(self, lat, lon):
        raise NotImplementedError
    def projectedToGeo(self, y, x):
        raise NotImplementedError
    def griddedToGeo(self, row, col):
        raise NotImplementedError


class latlon_GridDef(GridDef):
    '''
    Performs transformations on the unprojected lat/lon grid
    
    Convention for this class is that the "projected" coordinates
    are simply lat and lon.  The gridded coordinates
    determine cellsize by xCell and yCell, and place the ll corner
    at the lat lon specified by yOrig and xOrig, respectively.
    
    Cautions:
    - Only use -180 to 180 or 0 to 360.  Do not mix
    - Only use radians or degrees.  Do not mix.
    - Make sure parameters units and standard match
      coordinates being put in.
    '''
    @staticmethod
    def parm_list():
        return ['xOrig', 'yOrig', 'xCell', 'yCell', 'nRows', 'nCols']

    @staticmethod
    def requiredParms():
        '''Parameters that must be in the dictonary passed to instantiate'''
        return {"xOrig":('The longitude of the lower-left corner '\
                             'of the domain', 'decimal'),
                "yOrig":('The latitude of the lower-left corner '\
                             'of the domain', 'decimal'),
                "xCell":('The size of a gridcell in the x '\
                             '(longitude) direction in degrees', 'posdecimal'),
                "yCell":('The size of a gridcell in the y '\
                             '(latitude) direction in degrees', 'posdecimal'),
                "nRows":('The number of rows in the grid', 'posint'),
                "nCols":('The number of columns in the grid', 'posint')}
    def __init__(self, parmDict):
        GridDef.__init__(self, parmDict)
        
        # even though IO interface handles casting already,
        # a catchblock has been added here for safety
        # in case someone wants to use this class directly
        castDict = {"xOrig":float, "yOrig":float,
                    "xCell":float, "yCell":float,
                    "nRows":int, "nCols":int}
        for (k,func) in castDict.items():
            parmDict[k] = func(parmDict[k])
        self.parms = parmDict
    def indLims(self):
        return (0, self.parms['nRows']-1, 0, self.parms['nCols']-1)
    def geoToProjected(self, lat, lon):
        (y, x) = (lat, lon)
        return (y, x)
    def geoToGridded(self, lat, lon):
        row = (lat-self.parms['yOrig'])/self.parms['yCell']
        col = (lon-self.parms['xOrig'])/self.parms['xCell']
        return (row, col)
    def projectedToGeo(self, y, x):
        (lat, lon) = (y, x)
        return (lat, lon)
    def griddedToGeo(self, row, col):
        lat = row*self.parms['yCell']+self.parms['yOrig']
        lon = col*self.parms['xCell']+self.parms['xOrig']
        return (lat, lon)

    
class lcc2par_GridDef(GridDef):
    '''Performs transformations with the lambert conic conformal'''
    @staticmethod
    def parm_list():
        return ['stdPar1', 'stdPar2', 'refLat', 'refLon', 'xOrig', 'yOrig', 
                'xCell', 'yCell', 'nRows', 'nCols', 'earthRadius']

    @staticmethod
    def requiredParms():
        '''
        parameters that must be in the dictionary passed to
        instantiate this class
        '''
        return {"stdPar1":('One of the 2 standard parallels (latitudes) '\
                               'used to define the Lambert Conic '\
                               'Conformal projection, in degrees', 'decimal'),
                "stdPar2":('The second of the standard parallels (latitudes) '\
                               'used to define the Lambert Conic '\
                               'Conformal projection, in degrees.  '\
                               'Set this equal to stdPar1 if the '\
                               'single-parallel form of the projection '\
                               'is being used', 'decimal'),
                "refLat":('The reference latitude upon which the projection '\
                              'is centered, in degrees.  This is the '\
                              'YCENT value in the GRIDDESC file', 'decimal'),
                "refLon":('The reference longitude upon which the projection '\
                              'is centered, in degrees.  This is BOTH the '\
                              'XCENT and PROJ_GAMMA values in the GRIDDESC '\
                              'file.  If these values are not identical, '\
                              'do not use this function', 'decimal'),
                "xOrig":('The location of the origin in projected '\
                             'x coordinates, in the same units as earthRadius.'\
                             '  This is the XORIG value in the GRIDDESC file', 
                         'decimal'),
                "yOrig":('The location of the origin in projected y '\
                             'coordinates, in the same units as earthRadius.  '\
                             'This is the YORIG value in the GRIDDESC file', 
                         'decimal'),
                "xCell":('The x dimension of a cell in projected coordinates, '\
                             'in the same units as earthRadius.  This is the '\
                             'XCELL value in the GRIDDESC file', 'posdecimal'),
                "yCell":('The y dimension of a cell in projected coordinates,'\
                             ' in the same units as earthRadius.  This is the '\
                             'YCELL value in the GRIDDESC file', 'posdecimal'),
                "nRows":('The number of rows in the grid', 'posint'),
                "nCols":('The number of columns in the grid', 'posint'),
                "earthRadius":('The assumed radius of the Earth '\
                                  '(assumed spherical).  Must match units used'\
                                  ' for xCell and yCell','posdecimal')}
    def __init__(self, parms):
        GridDef.__init__(self, parms)

        # cast everything even though type is ensured (even though IO
        # interface does this) in case someone wants to use the class directly
        castDict = {"stdPar1" : float, "stdPar2" : float,
                    "refLat" : float, "refLon" : float,
                    "xOrig" : float, "yOrig" : float,
                    "xCell" : float, "yCell" : float,
                    "nRows" : int, "nCols" : int}
        for (k,func) in castDict.items():
            parms[k] = func(parms[k])
        self.parms = parms

        # remap parms into a dictionary readable by pyproj
        pyprojParms = {'proj'   : 'lcc',
                       'a'      : parms['earthRadius'],
                       'lat_1'  : parms['stdPar1'],
                       'lat_2'  : parms['stdPar2'],
                       'lat_0'  : parms['refLat'],
                       'lon_0'  : parms['refLon'],
                       'x_0'    : 0,
                       'y_0'    : 0}
        self.__proj = Proj(pyprojParms)
    def indLims(self):
        return (0, self.parms['nRows']-1,
                 0, self.parms['nCols']-1)
    def geoToProjected(self, lat, lon):
        (x,y) = self.__proj(lon, lat)
        return (y,x)
    def geoToGridded(self, lat, lon):
        (y,x) = self.geoToProjected(lat, lon)
        # transform x and y to new origin
        y = y-self.parms['yOrig']
        x = x-self.parms['xOrig']
        # convert from projected coordinates to cells
        row = y/self.parms['yCell']
        col = x/self.parms['xCell']
        return (row, col)
    def projectedToGeo(self, y, x):
        (lon,lat) = self.__proj(x, y, inverse=True)
        return (lat, lon)
    def griddedToGeo(self, row, col):
        y = row*self.parms['yCell']+self.parms['yOrig']
        x = col*self.parms['xCell']+self.parms['xOrig']
        return self.projectedToGeo(y,x)        



'''
Module for utility functions useful in a variety
of GIS programs

@author: Jacob Oberman
'''

import time
import calendar
from itertools import izip

import numpy
import netCDF4

import filetypes

def wrap_lon_0_360(lon):
    '''Wrap a single longitude to the interval [0, 360)'''
    while lon < 0:
        lon += 360
    while lon >= 360:
        lon -= 360
    return lon

def wrap_lon_neg180_180(lon):
    '''Wrap a single longitude to the interval (-180, 180]'''
    while lon <= -180:
        lon += 360
    while lon > 180:
        lon -= 360
    return lon

def timestr_to_nsecs(timestr, 
                     epoch='00:00:00 01-01-1970', 
                     format='%H:%M:%S %m-%d-%Y'):
    '''
    Convert a given string to the number of seconds
    since a specified epoch.
    
    Defaults to converting into Unix time
    
    Inputs:
        timestr - a string representing the time to convert
        epoch - a string representing desired epoch time
        format - a format string as in time.strptime used for timestr and epoch
        
    Outputs: 
        double - number of seconds since epoch
    '''
    desStructTime = time.strptime(timestr, format)
    epochStructTime = time.strptime(epoch, format)
    desUnixTime = calendar.timegm(desStructTime)
    epochUnixTime = calendar.timegm(epochStructTime)
    return desUnixTime - epochUnixTime

def nsecs_to_timestr(nSecsSinceEpoch,
                     epoch='00:00:00 01-01-1970',
                     format='%H:%M:%S %m-%d-%Y'):
    '''
    Convert the number of seconds since a specified epoch
    to a human-readable time string.
    
    Defaults to converting out of Unix time
    
    Inputs:
        nSecsSinceEpoch - number of seconds since specified epoch
        epoch - a string in the style of format representing epoch time
        format - a format string to convert into as in time.strptime
        
    Outputs:
        string - Human readable time string
    '''
    tStructAtEpoch = time.strptime(epoch, format)
    nSecsAtEpoch = calendar.timegm(tStructAtEpoch)
    nSecsUnix = nSecsSinceEpoch + nSecsAtEpoch
    tStructUnix = time.gmtime(nSecsUnix)
    return time.strftime(format, tStructUnix)
    
def UTCoffset_from_lon(lon):
    '''
    Calculate the approximate offset from UTC based on longitude.
    Offset is returned in seconds and will be positive
    or negative as appropriate
    
    Does not account for daylight savings time.  Uses an 
    algorithm based on simply the longitude, and therefore
    does not account for actual political timezones.
    
    Longitude assumed to be in degrees.
    '''
    hours2secs = 60*60
    return hours2secs*round(wrap_lon_neg180_180(lon)/15.0)

def find_occurences(superArray, subArray):
    '''
    Find the occurrences of a particular subarray within a superarray, 
    searching along the rightmost axis of the array.  Returns a logical array
    with of rank 1 less than the superarray.  Output is true where the subarray
    matched, false otherwise.
    '''
    return numpy.apply_along_axis(numpy.array_equal, -1, superArray, subArray)

def write_grid_to_netcdf(griddef, outFname):
    '''
    Function to create netCDF files that contain
    the lat/lon data needed to plot for a given grid 
    definition that uses rows/columns
    
    Inputs:
        griddef - an instantiated griddef object
        outFname - a path to the outfile (will be clobbered)
    '''
    (minRow, maxRow, minCol, maxCol) = griddef.indLims()
    nRows = maxRow-minRow+1
    nCols = maxCol-minCol+1
    # create the index vectors we need to make grids
    (cols, rows) = numpy.meshgrid(numpy.arange(minCol, maxCol+1), 
                                  numpy.arange(minRow, maxRow+1))
    cols = cols.astype(numpy.float32) # cast as precaution
    rows = rows.astype(numpy.float32)
    # write out netcdf
    fid = netCDF4.Dataset(outFname, 'w', format='NETCDF3_CLASSIC')
    # create dimensions and variables
    fid.createDimension('row', nRows)
    fid.createDimension('col', nCols)
    dims = ('row', 'col')
    # create the 5 grid definitions
    offsets = [(0,0), (1,0), (1,1), (0,1), (.5,.5)] # row,col
    labels = ['ll', 'ul', 'ur', 'lr', 'cent']
    for (lbl, (rowOff, colOff)) in izip(labels,offsets):
        lon = fid.createVariable(lbl+'_lon', 'f', dims)
        setattr(lon, 'Units', 'degrees_east')
        lat = fid.createVariable(lbl+'_lat', 'f', dims)
        setattr(lat, 'Units', 'degrees_north')
        (lat[:], lon[:]) = griddef.griddedToGeo(rows+rowOff, cols+colOff)
    # write grid parameters to file as global attributes
    setattr(fid, 'Projection', griddef.__class__.__name__[:-8])
    for (k,v) in griddef.parms.iteritems():
        setattr(fid, k, v)
    fid.close()

def parse_fromFile_input_file(inFileName, dryRun):
    '''
    Open and read the file and parse it 
    for various parameter values.
    Generate and return a command line call
    that can be used to call whips
    with those parameter values.
    '''
    call = []
    attrs = []
    f = open(inFileName, 'r')
    s = f.readline()
    try:
        while(s != ""):
            if(s == "BEGIN\n"):
                while(s != ""):
                    s = f.readline()
                    if(s == "END\n"):
                        return call + ["--projAttrs"] + attrs
                    if(s[0] == '.'):
                        continue
                    s = s.split('"')
                    if(len(s) > 1 and dryRun):
                        print "Warning: Found line containing "\
                              "stray quotation marks... stripping."
                    words = ("".join(s)).split()
                    '''Determine what the line is supposed to do'''
                    if(words == []):
                        continue
                    elif(words[1] != "="):
                        break
                    elif(words[0] == "DIRECTORY"):
                        call += ["--directory", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "FILELIST"):
                        call += ["--fileList"] + words[2:]
                    elif(words[0] == "FILETYPE"):
                        call += ["--filetype", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "GRIDPROJ"):
                        call += ["--gridProj", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "MAPFUNC"):
                        call += ["--mapFunc", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "OUTFUNC"):
                        call += ["--outFunc",
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "OUTDIRECTORY"):
                        call += ["--outDirectory", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "OUTFILENAME"):
                        call += ["--outFileName", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "INCLUDEGRID"):
                        call += ["--includeGrid",
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "VERBOSE"):
                        call += ["--verbose", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(words[0] == "INTERACTIVE"):
                        call += ["--interactive", 
                                 "{0}".format(' '.join(words[2:]))]
                    elif(dryRun):
                        attrs += ['"{0}:{1}"'.format(words[0], \
                                                         ' '.join(words[2:]))]
                    else:
                        attrs += ['{0}:{1}'.format(words[0], \
                                                         ' '.join(words[2:]))]
                break
            s = f.readline()
    except:
        pass
    if(s != ""):
        raise SyntaxError ("Invalid input file.  File must be formatted " \
                      "correctly.\nCheck line '{0}' and try again".format(s))
    raise SyntaxError ("Invalid input file.  Check that the file matches the "\
                      "format described in the documentation and try again")


def parse_filetype(namespace):
    '''
    Open and read the filetype file
    to add associated parameters to namespace 
    '''
    filetype = getattr(filetypes, namespace.filetype + "_filetype")
    printfiletype = namespace.filetype
    wng = ""

    for attr in dir(filetype):
        if attr == "parser":
            setattr(namespace, "filetype", getattr(filetype, attr))
        elif attr == "doutf":
            setattr(namespace, "outFunc", getattr(filetype, attr))
        elif attr[0] == "_":
            pass
        else:
            try:
                wng += "Warning: Value {0} supplied for attribute {1} " \
                      "has been ignored. Filetype {2} does not support "\
                      "custom values for this parameter.\n".format(\
                      getattr(namespace, attr), attr, printfiletype)
            except AttributeError:
                pass
            setattr(namespace, attr, getattr(filetype, attr))
            wng += "Value for {0} given by filetype information\n".format(attr)

    try:
        if (namespace.verbose != 'False'):
            print wng
    except:
        pass
    return namespace
